from typing import List, Optional
from pydantic import BaseModel
from socket import AF_INET, SOCK_DGRAM, SOL_SOCKET, SO_BROADCAST, socket
import asyncio
import json
from fakts.discovery.base import FaktsEndpoint, Beacon
from koil import Koil, unkoil
from koil.composition import KoiledModel
import logging


logger = logging.getLogger(__name__)


class BeaconProtocol(asyncio.DatagramProtocol):
    pass


class Binding(BaseModel):
    interface: str
    broadcast_addr: str
    bind_addr: str
    broadcast_port: int = 45678
    magic_phrase: str = "beacon-fakts"


def retrieve_bindings() -> List[Binding]:
    """Uses the netifaces library to retrieve all available interfaces and
    if they are up and running and have a broadcast address, it will return
    a list of bindings for the beacon to use.

    Raises:
        ImportError: An importError is raised if the netifaces library is not installed

    Returns:
        List[Binding]: The list of bindings
    """

    try:
        import netifaces
    except ImportError as e:
        raise ImportError(
            "netifaces is required to use the advertised discovery. please install it seperately or install fakts with the 'beacon' extras"
        ) from e

    potential_bindings: List[Binding] = []

    for interface in netifaces.interfaces():
        addrs = netifaces.ifaddresses(interface)
        if netifaces.AF_INET in addrs:
            informations = addrs[netifaces.AF_INET]
            for i in informations:

                if "broadcast" in i:
                    potential_bindings.append(
                        Binding(
                            interface=interface,
                            bind_addr=i["addr"],
                            broadcast_addr=i["broadcast"],
                        )
                    )
    return potential_bindings


async def advertise(
    binding: Binding,
    endpoints: List[Beacon],
    interval: int = 1,
    iterations: int = 10,
) -> None:
    """Advertises the given endpoints on the given binding

    This function opens a udp socket and sends the endpoints as json to the broadcast address
    on the given port. It will repeat this for the given number of iterations with the given
    interval in between.

    If interval is -1 it will repeat forever, until this task is cancelled

    Args:
        binding (Binding): The binding to use (interface, broadcast address, port)
        endpoints (List[FaktsEndpoint]): The list of endpoints to advertise
        interval (int, optional): The interval between a beacon send in seconds. Defaults to 1.
        iterations (int, optional): The amount of sends that should happen, -1 means infinite (until cancelled). Defaults to 10.

    """

    s = socket(AF_INET, SOCK_DGRAM)  # create UDP socket.

    try:
        s.bind((binding.bind_addr, 0))
        s.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)  # this is a broadcast socket

        loop = asyncio.get_event_loop()
        transport, pr = await loop.create_datagram_endpoint(BeaconProtocol, sock=s)

        messages = [
            bytes(binding.magic_phrase + json.dumps(beacon.dict()), "utf8")
            for beacon in endpoints
        ]
        i = 1
        while i <= iterations or iterations == -1:
            for message in messages:
                transport.sendto(
                    message, (binding.broadcast_addr, binding.broadcast_port)
                )
                logger.debug(f"Send Message {message}")

            await asyncio.sleep(interval)
            i += 1
    except asyncio.CancelledError as e:
        transport.close()
        s.close()
        raise e




import asyncio
import contextvars
import logging
from typing import Any, Dict, Set

from koil.composition import KoiledModel
from koil.helpers import unkoil
from pydantic import Field

from fakts.errors import FaktsError, GroupNotFound
from fakts.grants.base import FaktsGrant

logger = logging.getLogger(__name__)
current_fakts: contextvars.ContextVar["Fakts"] = contextvars.ContextVar("current_fakts")


class Fakts(KoiledModel):
    """Fakts is any asynchronous configuration loader.

    Fakts provides a way to concurrently load and access configuration from different
    sources in async and sync environments.

    It is used to load configuration from a grant, and to access it in async
    and sync code.

    A grant constitutes the way to load configuration. It can be a local config file
    (eg. yaml, toml, json), environemnt variables, a remote configuration (eg. from
    a fakts server) a database, or any other source.  It will be loaded either on
    call to `load`,  or on  a call to `get` (if auto_load is set to true).

    Additionaly you can compose grants with the help of meta grants in order to
    load configuration from multiple sources.

    Example:
        ```python
        async with Fakts(grant=YamlGrant("config.yaml")) as fakts:
            config = await fakts.aget("group_name")
        ```

        or

        ```python
        with Fakts(grant=YamlGrant("config.yaml")) as fakts:
            config = await fakts.get("group_name")
        ```

    Fakts should be used as a context manager, and will set the current fakts context
    variable to itself, letting you access the current fakts instance from anywhere in
    your code (async or sync). To understand how the async sync code access work,
    please check out the documentation for koil.


    Example:
        ```python
        async with Fakts(grant=FailsafeGrant(
            grants=[
                EnvGrant(),
                YamlGrant("config.yaml")
            ]
        )) as fakts:
            config = await fakts.get("group_name")
        ```
        In this example fakts will load the configuration from the environment
        variables first, and if that fails, it will load it from the yaml file.


    """

    grant: FaktsGrant
    """The grant to load the configuration from"""

    hard_fakts: Dict[str, Any] = Field(default_factory=dict, exclude=True)
    """Hard fakts are fakts that are set by the user and cannot be overwritten by grants"""

    loaded_fakts: Dict[str, Any] = Field(default_factory=dict, exclude=True)
    """The currently loaded fakts. Please use `get` to access the fakts"""

    assert_groups: Set[str] = Field(default_factory=set)
    """Asserted groups are groups that are asserted to be present in the fakts. If 
    they are not present, an error will be raised"""

    allow_auto_load: bool = Field(
        default=True, description="Should we autoload on get?"
    )
    """Should we autoload the grants on a call to get?"""

    load_on_enter: bool = False
    """Should we load on connect?"""
    delete_on_exit: bool = False
    """Should we delete on connect?"""

    _loaded: bool = False
    _lock: asyncio.Lock = None
    _fakts_path: str = ""

    async def aget(
        self,
        group_name: str,
        auto_load=True,
        force_refresh=False,
    ):
        """Get Config

        Gets the currently active configuration for the group_name, by loading it from
        the grant if it is not already loaded.

        Steps:
            1. Acquire lock
            2. If not yet loaded and auto_load is True, load
            4. Return groups fakts

        Args:
            group_name (str): The group name in the fakts
            auto_load (bool, optional): Should we autoload the configuration
                                        if nothing has been set? Defaults to True.
            force_refresh (bool, optional): Should we force a refresh of the grants.
                                            Grants can decide their own refresh logic?
                                            Defaults to False.

        Returns:
            dict: The active fakts
        """
        assert (
            self._lock is not None
        ), "You need to enter the context first before calling this function"
        async with self._lock:
            if not self.loaded_fakts or force_refresh:
                if self.allow_auto_load and auto_load:
                    await self.aload(force_refresh=force_refresh)
                else:
                    raise FaktsError(
                        "No loaded fakts and auto_load is False. Please load first."
                    )

        try:
            config = self._getsubgroup(group_name)
        except GroupNotFound as e:
            if self.allow_auto_load and auto_load:
                await self.aload(force_refresh=True)
                config = self._getsubgroup(group_name)
            else:
                raise e

        return config

    def _getsubgroup(self, group_name: str) -> Dict[str, Any]:
        """Get subgroup

        Protected function to get a subgroup from the loaded fakts

        Args:
            group_name (str): The name of the group

        Raises:
            GroupNotFound: If the groups is not found in the loadedfakts

        Returns:
            Dict[str, Any]: The subgroups configuration as a dictioniary
        """
        config = {**self.loaded_fakts}

        for subgroup in group_name.split("."):
            try:
                config = config[subgroup]
            except KeyError as e:
                raise GroupNotFound(f"Could't find {subgroup} in fakts {config}") from e

        return config

    def has_changed(self, fakt_dict: Dict[str, Any], group: str):
        return (
            not fakt_dict or self._getsubgroup(group) != fakt_dict
        )  # TODO: Implement Hashing on config?

    async def arefresh(self):
        """Causes a Refresh on the grants. Grants can decide their own refresh logic."""
        self.loaded_fakts = None
        await self.aload(force_refresh=True)

    def get(self, *args, **kwargs):
        """Sync version of aget"""
        return unkoil(self.aget, *args, **kwargs)

    async def aload(self, force_refresh=False) -> Dict[str, Any]:
        """Loads the configuration from the grant and asserts the groups"""

        self.loaded_fakts = await self.grant.aload(force_refresh=force_refresh)
        if not self.assert_groups.issubset(set(self.loaded_fakts.keys())):
            raise GroupNotFound(
                f"Expected Configuration for {self.assert_groups - set(self.loaded_fakts.keys())}. "
            )

        self._loaded = True
        return self.loaded_fakts

    def load(self, **kwargs):
        """Sync version of aloads"""
        return unkoil(self.aload, **kwargs)

    async def __aenter__(self):
        current_fakts.set(self)
        self._lock = asyncio.Lock()
        if self.load_on_enter:
            await self.aload()
        return self

    async def __aexit__(self, *args, **kwargs):
        current_fakts.set(None)

    def _repr_html_inline_(self):
        return f"<table><tr><td>grant</td><td>{self.grant.__class__.__name__}</td></tr></table>"

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True
        json_encoders = {
            FaktsGrant: lambda x: f"Fakts Grant {x.__class__.__name__}",
        }


def get_current_fakts() -> Fakts:
    return current_fakts.get()



from typing import List
from fakts.grants.base import FaktsGrant
from fakts.grants.errors import GrantError
import logging

logger = logging.getLogger(__name__)


class FailsafeGrant(FaktsGrant):
    """A failsafe grant that tries to load a grant and if it fails it tries the next one. Only fails if all grants fail"""

    grants: List[FaktsGrant]

    async def aload(self, **kwargs):
        for grant in self.grants:
            try:
                config = await grant.aload(**kwargs)
                return config
            except GrantError as e:
                logger.exception(f"Failed to load {grant}", exc_info=True)
                continue

        raise GrantError("Failed to load any grants")



try:
    import tomllib as toml
except ImportError:
    try:
        import toml
    except ImportError:
        raise ImportError(
            "You need to install the `toml` package to use the toml grant."\
            "Or use python 3.11 and higher which comes with the `tomllib` package"
        )
from fakts.grants.base import FaktsGrant


class TomlGrant(FaktsGrant):
    filepath: str

    async def aload(self, force_refresh=False, **kwargs):
        with open(self.filepath, "r") as file:
            config = toml.load(file)

        return config



from fakts.grants.base import FaktsGrant
import os
from typing import Any, Dict, Optional
import pydantic
import datetime
import logging
import json

logger = logging.getLogger(__name__)


class CacheFile(pydantic.BaseModel):
    """Cache file model"""

    config: Dict[str, Any]
    created: datetime.datetime
    hash: str = ""


class CacheGrant(FaktsGrant):
    """Grant for caching data, caches the data of the its child grant in a file,
    if that file exists, and it is not expired, it will be used instead of delegating
    to the child grant."""

    grant: FaktsGrant = pydantic.Field(..., description="The grant to cache")
    cache_file: str = ".fakts_cache.json"
    hash: str = pydantic.Field(
        default_factory=lambda: "",
        description="Validating against the hash of the config",
    )
    skip_cache: bool = False
    expires_in: Optional[int]

    async def aload(self, force_refresh: bool = False):
        cache = None

        if os.path.exists(self.cache_file):
            with open(self.cache_file, "r") as f:
                x = json.load(f)
                try:
                    cache = CacheFile(**x)

                    if self.expires_in:
                        if (
                            cache.created + datetime.timedelta(seconds=self.expires_in)
                            < datetime.datetime.now()
                        ):
                            cache = None

                    if self.hash and cache.hash != self.hash:
                        cache = None

                except pydantic.ValidationError as e:
                    logger.error(f"Could not load cache file: {e}. Ignoring it")

        if cache is None or force_refresh or self.skip_cache:
            logger.info("Loading data from grant")
            data = await self.grant.aload(force_refresh=force_refresh)
            cache = CacheFile(
                config=data, created=datetime.datetime.now(), hash=self.hash
            )

        with open(self.cache_file, "w+") as f:
            json.dump(json.loads(cache.json()), f)

        return cache.config



from fakts.discovery.base import FaktsEndpoint
import aiohttp
import logging
import ssl
from fakts.discovery.errors import DiscoveryError
from typing import Optional, List

logger = logging.getLogger(__name__)

async def check_wellknown(url: str, ssl_context: ssl.SSLContext, timeout = 4) -> FaktsEndpoint:

    url = f"{url}.well-known/fakts"

    async with aiohttp.ClientSession(
        connector=aiohttp.TCPConnector(ssl=ssl_context),
        headers={"User-Agent": "Fakts/0.1", "Accept": "application/json"},
    ) as session:
        async with session.get(
            url,
            timeout=timeout,
        ) as resp:

            if resp.status == 200:
                data = await resp.json()

                if "name" not in data:
                    logger.error(f"Malformed answer: {data}")
                    raise Exception("Malformed Answer")

                return FaktsEndpoint(**data)


            else:
                logger.error(f"Could not retrieve on the endpoint: {resp.status}")
                raise DiscoveryError(f"Error! We could not retrieve the endpoint. {url} ")



async def discover_url( url: str, ssl_context: ssl.SSLContext, auto_protocols: Optional[List[str]] = None, allow_appending_slash: bool = False, timeout= 4) -> FaktsEndpoint:
    


    if "://" not in url:
            logger.info(f"No protocol specified on {url}")
            if not auto_protocols or len(auto_protocols) == 0:
                raise DiscoveryError("No protocol specified and no auto protocols specified")
            
            errors = []
            
            for protocol in auto_protocols:
                logger.info(f"Trying to connect to {protocol}://{url}")
                try:

                    if allow_appending_slash and not url.endswith("/"):
                        url = f"{url}/"


                    return await check_wellknown(f"{protocol}://{url}", ssl_context, timeout=timeout)
                except Exception as e:
                    logger.info(f"Could not connect to {protocol}://{url}")
                    errors.append((protocol, e))
                    continue

            errors_string = "\n".join([f"- {protocol}://{url}\n  " + str(e) for protocol, e in errors])

            raise DiscoveryError(f"Could not connect via any protocol: \n{errors_string}")

    if allow_appending_slash and not url.endswith("/"):
        url = f"{url}/"

    return await check_wellknown(url, ssl_context, timeout=timeout)